# (c) Copyright QuantumWise A/S
# See the file "LICENSE" in the AddOns directory of the ATK distribution
# for the full license governing this code.
# This code originally from Stefano Borini's wavemol core/fileaccess modules, released under BSD License

import re

# Regexp patterns used to parse XYZ files.
FLOAT_RE_STR = r'[-+]?[0-9]+[.]?[0-9]*(?:[eE][-+]?[0-9]+)?'
WORD_RE_STR  = r'[a-zA-Z]*'
INT_RE_STR   = r'\d+'
ELEMENT_RE_STR = r'[A-Z]+[a-z]*[a-z]*'

NUM_ATOMS_RE = re.compile(r"^\s*("+INT_RE_STR+")\s*$")
# Can be without spaces between coords.
# The minus sign may act as a separator for very ugly formatters.
ATOM_ROW_RE = re.compile("^\s*("+ELEMENT_RE_STR+")\s+("+FLOAT_RE_STR+")\s*("+FLOAT_RE_STR+")\s*("+FLOAT_RE_STR+")*")
COMMENT_RE = re.compile("^(.*)$")


class XYZFileRawReader:
    """
    Parses a XYZ file in a raw way. Raw means that it does not add any intelligence to the
    returned values, except for basic format checking. It does not check validity of element
    labels, nor returns advanced objects
    """
    def __init__(self, filename):

        # Save the filename.
        self._filename = filename
        # A list of dictionaries, one for each molecule. Lightweight and easy to print.
        self._molecules = []

        # Parse the data.
        self._parse()

    def _parse(self):
        """
        Performs the parsing of the file specified in the constructor.
        Called by the constructor.
        """
        # Save the filename.
        filename = self._filename

        # Setup the State machine states.
        current_molecule = None
        STATE_NUM_ATOMS, STATE_COMMENT, STATE_ATOM_ROW = range(3)
        state = STATE_NUM_ATOMS

        # Open the file.
        with open(self._filename, "r") as f:
            for line_num, line in enumerate(f):
                # Skip empty lines and comments.
                if (not state == STATE_COMMENT) and (len(line.strip()) == 0 or line.strip()[0] == "#"):
                    continue

                #
                if state == STATE_NUM_ATOMS:
                    m = NUM_ATOMS_RE.match(line)
                    if m is not None:
                        # We match a new single integer line, a new molecule starts.
                        # validate the old one and start a new one
                        current_molecule = {"num_atoms": int(m.group(1)),
                                            "comment":   None,
                                            "atom_list": []}
                        state = STATE_COMMENT
                        # Next line.
                        continue
                    else:
                        raise IOError("Invalid parse, line %d, file %s. " % (line_num, filename))
                #
                elif state == STATE_COMMENT:
                    current_molecule["comment"] = line.strip()
                    state = STATE_ATOM_ROW
                    continue
                #
                elif state == STATE_ATOM_ROW:
                    m = ATOM_ROW_RE.match(line)
                    if m is not None:
                        element = m.group(1)
                        coords = (float(m.group(2)), float(m.group(3)), float(m.group(4)))
                        current_molecule["atom_list"].append({"element": element,
                                                              "coords": coords})
                        state = STATE_ATOM_ROW

                        continue

                    m = NUM_ATOMS_RE.match(line)
                    if m is not None:
                        # Old molecule is finished, new molecule starts.
                        if current_molecule["num_atoms"] != len(current_molecule["atom_list"]):
                            raise IOError("Invalid parse, line %d, file %s. " % (line_num, filename))
                        self._molecules.append(current_molecule)
                        current_molecule = {"num_atoms": int(m.group(1)),
                                            "comment":   None,
                                            "atom_list": []}
                        state = STATE_COMMENT
                        # Next line.
                        continue

        #
        if current_molecule:
            if current_molecule["num_atoms"] != len(current_molecule["atom_list"]):
                raise IOError("At line %d in %s: found %d atoms, %d expected." % (line_num, filename, len(current_molecule["atom_list"]), current_molecule["num_atoms"]))
            self._molecules.append(current_molecule)

    def comment(self, molecule=0):
        """
        Returns the string comment for a given molecule, or the first molecule if no parameter is specified

        @param molecule : the index of the molecule
        @type           : int
        @return A string containing the comment line, stripped of end of line spaces

        """

        if len(self._molecules) == 0 or not (0 <= molecule < len(self._molecules)):
            raise IndexError("Invalid molecule index")

        return self._molecules[molecule]["comment"]

    def numOfMolecules(self):
        """
        Returns the number of molecules available in the file

        @return An integer containing the number of molecules
        """
        return len(self._molecules)

    def numOfAtoms(self, molecule=0):
        """
        Returns the number of atoms in the specified molecule (or the first molecule if unspecified)

        @param molecule : The index of the molecule (default 0).
        @type           : int

        @return An integer with the number of atoms in that molecule
        """
        if len(self._molecules) == 0 or not (0 <= molecule < len(self._molecules)):
            raise IndexError("Invalid molecule index")

        return len(self._molecules[molecule]["atom_list"])

    def atomList(self, molecule=0):
        """
        Returns a list of dicts, each dict with two keys:
            - "element": A string of the element symbol.
            - "coords":  A tuple of three floating point values, for x, y, z.

        @param molecule : The index of the molecule.
        @type           : int.

        @return A list as specified
        """
        return self._molecules[molecule]["atom_list"]
